package token;

import static recursion.TailCalls.done;
import static recursion.TailCalls.tailCall;
import static sexpression.Nil.NIL;

import java.util.function.Supplier;

import file.FilePosition;
import recursion.TailCall;
import sexpression.Cons;
import sexpression.SExpression;

public abstract class Token {

    private String text;
    private FilePosition position;

    public Token(String text, FilePosition position) {
        this.text = text;
        this.position = position;
    }

    public String getText() {
        return text;
    }

    public FilePosition getPosition() {
        return position;
    }

    // --------------------------------------------------------------------------------------------
    // s-expression
    // ::= NUMBER | IDENTIFIER | STRING | QUOTE_MARK s-expression | LEFT_PAREN s-expression-tail
    //
    // s-expression-tail
    // ::= RIGHT_PAREN | s-expression s-expression-tail
    // --------------------------------------------------------------------------------------------

    public abstract SExpression parseSExpression(Supplier<Token> getNextToken);

    public SExpression parseSExpressionTail(Supplier<Token> getNextToken) {
        Cons backwardsList = parseListBackwards(NIL, getNextToken).invoke();

        return reverse(NIL, backwardsList).invoke();
    }

    public TailCall<Cons> parseListBackwards(Cons accumulator, Supplier<Token> getNextToken) {
        Cons nextAccumulator = new Cons(parseSExpression(getNextToken), accumulator);
        Token next = getNextToken.get();

        return tailCall(() -> next.parseListBackwards(nextAccumulator, getNextToken));
    }

    // TODO - refactor into a built in function?
    private TailCall<SExpression> reverse(Cons accumulator, Cons list) {
        if (list.isNull())
            return done(accumulator);

        Cons nextAccumulator = new Cons(list.getFirst(), accumulator);
        Cons rest = (Cons) list.getRest();

        return tailCall(() -> reverse(nextAccumulator, rest));
    }

}
